// ###############################
// # Gecko Custom Commands (GCC) #
// ###############################

import { world } from "@minecraft/server";

function _notify(msg = "Specify a message", type = "msg") {
    if (type == "err") {
        world.sendMessage("[§cERROR§r] " + msg)
    }
    else {
        world.sendMessage(msg)
    }
}

function _splitAndCapitalize(word) {
    // Split the word at capital letters using regular expression
    let splitWords = word.split(/(?=[A-Z])/);
    // Capitalize the first word
    splitWords[0] = splitWords[0].charAt(0).toUpperCase() + splitWords[0].slice(1);
  
    return splitWords.join(" ");
}

/**
 * The blueprint for a custom command
 * @param name The name of the command, used as "!name".
 * @param arguments The names of all the arguments (used by the !help command). 
 *                  If the last argument ends with "+", arguments are expected to be able to be unlimited
 * @param title The title to be used int !help. If non is provided, use first name
 * @param description An optional description of what the command does. The !help command will display this
 * @method execute The function that the command needs to trigger. The arguments (if any) will be passed to this function
 */
export class GGcommand {
    static name = "";
    static arguments = [];
    static title = null;
    static description = null;

    static execute() {
        throw new Error('Method "execute" must be implemented by subclass.');
    }
}

// Built in help command, giving an explanation of all of the commands
class cmd_Help extends GGcommand {
    static name = ["help", "h", "?"]
    static description = "This command will show you all of the commands and how to write them"

    static execute() {
        _notify("§eHere's all of the commands supported at the moment:")
        for (let i = 0; i < _cmdManager.commands.length; i++) {
            const name = _cmdManager.commands[i].name
            _notify("\n§9" + _splitAndCapitalize(_cmdManager.commands[i].title ?? _cmdManager.commands[i].name[0]) + "§r")
            _notify(_cmdManager.prefix + (Array.isArray(name) ? name.join(" / !") : name) + " " + _cmdManager.commands[i].arguments.join(" "));
            if (_cmdManager.commands[i].description != null) {
                _notify(_cmdManager.commands[i].description)
            }
        }
    }
}

/**
 * add multiple commands to the possible commands
 * @param commands An array of classes inheriting from GGcommand
 */
export function addCommands(commands = []) {
    for (let cmd of commands) {
        _cmdManager.commands.push(cmd)
    }
}

world.beforeEvents.chatSend.subscribe((eventData) => {
    // Checks if a command has been types in chat, if so it runs matchCommand to locate and trigger the appropriate command
    if (eventData.message.substr(0, 1) == _cmdManager.prefix) {
        eventData.cancel = true
        try {
            _cmdManager.matchCommand(_cmdManager.stringToCommand(eventData.message), eventData.sender)
        }
        catch (err) {
            _notify("Command not executed: " + err.message, "err")
        }
    }
});



const _cmdManager = {
    prefix: "!",
    commands: [cmd_Help],

    // Converts a simple string to an array
    // Each array element will be a command argument
    stringToCommand: function (_strCmd) {
        // Removes prefix
        let strCmd = _strCmd.substr(1, _strCmd.length).trim();

        // Groups arguments that were in brackets into a single array element
        let command = strCmd.split(" "), i = 0, begin, end;

        function findNext(chr) {
            while (i < command.length) {
                if (command[i].indexOf(chr) != -1) {
                    command[i] = command[i].replace(chr, "");
                    return (i);
                }
                i++;
            }
            return (-1);
        }

        do {
            end = -1;
            begin = findNext("[");
            if (begin != -1) {
                end = findNext("]");
                if (end != -1) {
                    let arr = command.splice(begin, end - begin + 1); // The new array argument
                    // remove any spaces and splits array entries at ","
                    // e.g. [0,1,2 ,3, 4] becomes [0,1,2,3,4]
                    arr = arr.join("");
                    arr = arr.split(",");
                    command.splice(begin, 0, arr);
                    i = begin + 1;
                }
            }
        } while (end != -1);

        return command;
    },

    // Goes through the commands array to find the approriate command and execute it
    matchCommand: function (cmd, sender) {
        let command = cmd
        for (let i = 0; i < this.commands.length; i++) {
            for (let j = 0; j < this.commands[i].name.length; j++) { //second loop loops through all names for a specific command
                if (command[0] == this.commands[i].name[j]) {
                    // Checks if too many arguments were passed into the command and throws an error
                    if (command.length - 1 > this.commands[i].arguments.length && !this.commands[i].arguments[this.commands[i].arguments.length - 1].endsWith("+")) {
                        _notify("Too many arguments passed to command", "err")
                        return
                    }
                    let args = command
                    // Removes the command name
                    args.shift()
                    // Runs the function that was specified by the command
                    this.commands[i].execute(args, sender)
                    return
                }
            }
        }
        _notify(command[0] + " is not a recognised command", "err")
    },
}