const fs = require("fs");

// stand in for "@minecraft/server" Vector class
class Vector {
    constructor(x, y, z) {
      this.x = x;
      this.y = y;
      this.z = z;
    }
  
    static add(vector1, vector2) {
        return new Vector(vector1.x + vector2.x, vector1.y + vector2.y, vector1.z + vector2.z);
    }
    
    static subtract(vector1, vector2) {
    return new Vector(vector1.x - vector2.x, vector1.y - vector2.y, vector1.z - vector2.z);
    }
}


function cone(pos, size) {
    let regionsToFill = []
    const maxHeight = (parseInt(size) + 1) * 2;

    for (let height = 0; height < maxHeight; height++) {
        if (height % 2 !== 0) {
            height++;
        }
        let currentRad = parseInt(size) + 1 - (height / 2);

        for (let tx = -currentRad; tx < currentRad + 1; tx++) {
            for (let tz = -currentRad; tz < currentRad + 1; tz++) {
                const distance = Math.sqrt(Math.pow(tx, 2) + Math.pow(tz, 2)) + 0.2;
                if (distance <= currentRad) {
                    const placePos = new Vector(pos.x + tx, pos.y + height, pos.z + tz);
                    regionsToFill.push(placePos);
                }
            }
        }
    }

    const condensedLines = condenseToLines(regionsToFill);
    const fullHeightRegions = condensedLines.map((coordinate) => {
        return {
            start: coordinate.start,
            end: Vector.add(coordinate.end, new Vector(0, 2, 0))
        }
    })
    return fullHeightRegions;
}


function cylinder(pos, size, minSizeForCenterCube=9) {
    let rad = parseInt(size) + 1;
    let regionsToFill = [];
    let outerRegions = []

    // If ball is large enough, the center is filled with a number of large cubes
    // This results in weird looking balls when the radius is too small however, so
    // only create these regions if the ball is big enough 
    if (rad >= minSizeForCenterCube) {
        const center = pos;

        // Fill the central part of the ball as a larger cube
        const centralCubeSize = Math.floor(rad * 0.59); // Adjust the size as needed
        const centralCubeStart = Vector.subtract(center, new Vector(centralCubeSize, 0, centralCubeSize));
        const centralCubeEnd = Vector.add(center, new Vector(centralCubeSize, 0, centralCubeSize));
        regionsToFill.push({ start: centralCubeStart, end: centralCubeEnd });

        // Create layered edges around central cube
        const startValues = centralCubeSize + 1;
        for (let i = startValues; i < rad; i++) {
            const placements = [
                { pos: new Vector(pos.x + i, pos.y, pos.z), axis: "xt" },
                { pos: new Vector(pos.x, pos.y, pos.z + i), axis: "zt" },

                { pos: new Vector(pos.x - i, pos.y, pos.z), axis: "xb" },
                { pos: new Vector(pos.x, pos.y, pos.z - i), axis: "zb" },
            ];

            placements.forEach(({ pos, axis }) => {
                const size = rad - centralCubeSize;
                const sizeDecrease = Math.floor(i - startValues - 1);
                const axisValues = {
                    "xt": [0, 0, size - sizeDecrease],
                    "zt": [size - sizeDecrease, 0, 0],

                    "xb": [0, 0, size - sizeDecrease],
                    "zb": [size - sizeDecrease, 0, 0]
                };

                const CubeStart = Vector.subtract(pos, new Vector(...axisValues[axis]));
                const CubeEnd = Vector.add(pos, new Vector(...axisValues[axis]));
                
                regionsToFill.push({ start: CubeStart, end: CubeEnd});
            });
        }
    }

    for(let tx = -rad; tx < rad + 1; tx++){
        for(let tz = -rad; tz < rad + 1; tz++){
            if(Math.sqrt(Math.pow(tx, 2) + Math.pow(tz, 2)) + 0.2 <= rad)
            {
                const placePos = Vector.add(pos, new Vector(tx, 0, tz));
                let insideRegion = false;
                for (const region of regionsToFill) {
                    if (isInsideRegion(placePos, region)) {
                        insideRegion = true;
                        break;
                    }
                    
                }
                if (!insideRegion)
                    outerRegions.push(placePos);
            }
        }
    }

    const condensedLines = condenseToLines(outerRegions);
    regionsToFill = regionsToFill.concat(condensedLines)
    return regionsToFill
}


function ball(pos, size, minSizeForCenterCube=9) {
    let rad = parseInt(size) + 1;
    let regionsToFill = [];
    let outerRegions = []

    // If ball is large enough, the center is filled with a number of large cubes
    // This results in weird looking balls when the radius is too small however, so
    // only create these regions if the ball is big enough 
    if (rad >= minSizeForCenterCube) {
        const center = pos;

        // Fill the central part of the ball as a larger cube
        const centralCubeSize = Math.floor(rad * 0.59); // Adjust the size as needed
        const centralCubeStart = Vector.subtract(center, new Vector(centralCubeSize, centralCubeSize, centralCubeSize));
        const centralCubeEnd = Vector.add(center, new Vector(centralCubeSize, centralCubeSize, centralCubeSize));
        regionsToFill.push({ start: centralCubeStart, end: centralCubeEnd });

        // Create layered edges around central cube
        const startValues = centralCubeSize + 1;
        for (let i = startValues; i < rad; i++) {
            const placements = [
                { pos: new Vector(pos.x, pos.y + i, pos.z), axis: "yt" },
                { pos: new Vector(pos.x + i, pos.y, pos.z), axis: "xt" },
                { pos: new Vector(pos.x, pos.y, pos.z + i), axis: "zt" },

                { pos: new Vector(pos.x, pos.y - i, pos.z), axis: "yb" },
                { pos: new Vector(pos.x - i, pos.y, pos.z), axis: "xb" },
                { pos: new Vector(pos.x, pos.y, pos.z - i), axis: "zb" },
            ];

            placements.forEach(({ pos, axis }) => {
                const size = rad - centralCubeSize;
                const sizeDecrease = Math.floor(i - startValues - 1);
                const axisValues = {
                    "xt": [0, size - sizeDecrease, size - sizeDecrease],
                    "yt": [size - sizeDecrease, 0, size - sizeDecrease],
                    "zt": [size - sizeDecrease, size - sizeDecrease, 0],

                    "xb": [0, size - sizeDecrease, size - sizeDecrease],
                    "yb": [size - sizeDecrease, 0, size - sizeDecrease],
                    "zb": [size - sizeDecrease, size - sizeDecrease, 0]
                };

                const CubeStart = Vector.subtract(pos, new Vector(...axisValues[axis]));
                const CubeEnd = Vector.add(pos, new Vector(...axisValues[axis]));
                
                regionsToFill.push({ start: CubeStart, end: CubeEnd});
            });
        }
    }
    
    // Calculate the rounded outer parts of the ball as smaller individual cubes
    // Only adds new positions to areas not inside existing "regionsToFill", so
    // only calculate outer edges if there are central cubes, or whole ball if
    // there is no existing regions
    for (let tx = -rad; tx < rad + 2; tx++) {
        for (let ty = -rad; ty < rad + 2; ty++) {
            for (let tz = -rad; tz < rad + 2; tz++) {
                if (Math.sqrt(Math.pow(tx, 2) + Math.pow(ty, 2) + Math.pow(tz, 2)) + 0.2 <= rad) {
                    const placePos = Vector.add(pos, new Vector(tx, ty, tz));
                    let insideRegion = false;
                    for (const region of regionsToFill) {
                        if (isInsideRegion(placePos, region)) {
                            insideRegion = true;
                            break;
                        }
                        
                    }
                    if (!insideRegion)
                        outerRegions.push(placePos);
                }
            }
        }
    }


    const condensedLines = condenseToLines(outerRegions);
    regionsToFill = regionsToFill.concat(condensedLines)
    return regionsToFill
}

// Combine smaller cubes into larger strips
function condenseToLines(coordinates) {
    if (coordinates.length < 2) {
      return [];
    }
  
    const lines = [{ start: coordinates[0], end: coordinates[0] }];
    let line_i = 0;
    for (let i = 0; i < coordinates.length; i++) {
        if (lines[line_i].end.x === coordinates[i].x && lines[line_i].end.y === coordinates[i].y) {
            lines[line_i].end = coordinates[i]
        }
        else {
            lines.push({start: coordinates[i], end: coordinates[i]})
            line_i++
        }
    }
  
    return lines;
  }

function isInsideRegion(point, region) {
    const startX = Math.min(region.start.x, region.end.x);
    const endX = Math.max(region.start.x, region.end.x);
    const startY = Math.min(region.start.y, region.end.y);
    const endY = Math.max(region.start.y, region.end.y);
    const startZ = Math.min(region.start.z, region.end.z);
    const endZ = Math.max(region.start.z, region.end.z);

    return (
        point.x >= startX && point.x <= endX &&
        point.y >= startY && point.y <= endY &&
        point.z >= startZ && point.z <= endZ
    );
}


let bakedShapes = {
    ball: {},
    cylinder: {},
    cone: {}
};
for (let i = 1; i < 21; i++) {
    bakedShapes.ball[i] = ball(new Vector(0, 0, 0), i)
    bakedShapes.cylinder[i] = cylinder(new Vector(0, 0, 0), i)
    bakedShapes.cone[i] = cone(new Vector(0, 0, 0), i)
}





const data = `export const bakedShapes = ${JSON.stringify(bakedShapes, null, 2)};`;

// Write the regions data to a JSON file
fs.writeFile("./scripts/baked/bakedShapes.js", data, (err) => {
    console.log("Regions data has been baked");
});