import { world, system, BlockPermutation, BlockStates } from "@minecraft/server";

export function notify(msg = "Specify a message", type = "msg") {
    if (type == "err") {
        world.sendMessage("[§cERROR§r] " + msg);
    }
    else if (type == "warn") {
        world.sendMessage("[§eWARNING§r] " + msg);
    }
    else {
        world.sendMessage("[§3World§6Fab§r] " + msg);
    }
}

export function runCommand(dim, cmd) {
    system.run(() => {
        dim.runCommand(cmd);
    });
}

export function setBlock(dim, location, block) {
    system.run(() => dim.fillBlocks(location, location, block));
}

export function fillBlocks(dim, location_0, location_1, block, fillOptions = {}) {
    system.run(() => dim.fillBlocks(location_0, location_1, block, fillOptions));
}

export function fillBlocksMasked(dim, location_0, location_1, blockPerm, mask) {
    let maskOptions;
    for (const m of mask) {
        if (m.blockData == null) {
            const states = Object.keys(BlockPermutation.resolve(m.block).getAllStates());
            const possibleStates = states.map(s => {
                const values = BlockStates.get(s).validValues;
                return {
                    state: s,
                    values: values
                };
            });
            const statesWithValues = generateStateCombinations(possibleStates);
            for (const combination of statesWithValues) {
                maskOptions = { matchingBlock: BlockPermutation.resolve(m.block, combination) };

                fillBlocks(dim, 
                           location_0,
                           location_1,
                           blockPerm,
                           maskOptions);
            }

            continue;
        }
        else {
            maskOptions = { matchingBlock: BlockPermutation.resolve(m.block, m.blockData) }
            fillBlocks(dim, 
                       location_0,
                       location_1,
                       blockPerm,
                       maskOptions);
        }
    }
}

function removeOuterQuotes(str) {
    if ((str.startsWith('"') && str.endsWith('"')) 
    ||(str.startsWith("'") && str.endsWith("'"))) {
        return str.slice(1, -1);
    } 
    else {
        return str;
    }
}

/**
 * 
 * @param vec vector to convert
 * @param separator substring to separate the vector parameters
 */
export function vectorToString(vec, separator = " ") {
    return `${vec.x}${separator}${vec.y}${separator}${vec.z}`
}

/**
 * Takes a string and convert to its true type, e.g. "23" to 23, "false" to false, etc
 * @param value string
 * @returns value as actual type, or original string
 */
export function convertToActualType(value) {
    // int
    if (!isNaN(value) && Number.isInteger(parseFloat(value)))
        return parseInt(value, 10); // Convert to integer

    // float
    if (!isNaN(value))
        return parseFloat(value); // Convert to float

    // boolean
    if (value.toLowerCase() === "true") 
        return true;
    else if (value.toLowerCase() === "false")
        return false;

    // string
    return value;
}

export function BlockStateStringToObject(stateStrings) {
    if (stateStrings === "defaultState")
        return null;

    let result = {};
    for (let s of stateStrings) { 
        let [key, value] = s.split("=");
        key = removeOuterQuotes(key);
        value = removeOuterQuotes(value);
        result[key] = convertToActualType(value);
    }
    return result;
}
  


export function formatObject(obj, nest=false) {
    let string = "";
    for (let k of Object.keys(obj)) {
        if (typeof obj[k] == "object")
            string += formatObject(obj[k], true);
        else {
            if (nest)
                string += ("\n" + "    " + obj[k].toString());
            else
                string += ("\n" + obj[k].toString());
        }
    }

    return string;
}

export function formatMaskArray(mask) {
    return mask.map((obj, index) => {
        const maskedIndex = `  Mask ${index}:`;
        const objString = Object.entries(obj).map(([key, value]) => `    ${key}: ${value}`).join('\n');
        return `${maskedIndex}\n${objString}`;
    }).join('\n\n');
}

// Function to generate all possible combinations of states with their values
export function generateStateCombinations(statesWithValues) {
    const combinations = [];
    
    const generate = (currentIndex, currentCombination) => {
        if (currentIndex === statesWithValues.length) {
            const combinationObject = {};
            for (const entry of currentCombination) {
                combinationObject[entry.state] = entry.value;
            }
            combinations.push(combinationObject);
            return;
        }

        const state = statesWithValues[currentIndex];
        const statesArr = state.values
        for (let i = 0; i < statesArr.length; i++) {
            const value = statesArr[i]
            currentCombination.push({ state: state.state, value: value });
            generate(currentIndex + 1, currentCombination);
            currentCombination.pop();
        }
    }

    generate(0, []);
    return combinations;
}
