import { bakedShapes } from "../baked/bakedShapes";
import { BlockStateStringToObject, fillBlocks, fillBlocksMasked, formatMaskArray, formatObject, notify } from "../other/helpers";
import { Vector, BlockPermutation, world } from "@minecraft/server";


export class Brush {
    /**
     * 
     * @param sender The player who is using the brush
     * @param shape The shape that the brush will generate
     * @param size The radius of the brush
     * @param block The block name
     * @param blockData Block State String describing the block states used
     * @param mask Null for no mask or Array of Objects for block name and data
     * @param silent if true, mute creation notification 
     */
    constructor(sender, shape, size, block, blockData, mask, silent=false) {
        this.owner = sender;
        this.shape;
        this.size;
        this.block;
        this.blockData;
        this.mask;

        this._firstTime = true;

        this.setShape(shape);
        this.setSize(size);
        if (block != null && blockData != null)
            this.setBlock(block, blockData);
        this.setMask(mask);

        if (!silent)
            notify(`A new §9${this.blockPerm.type.id}§r §2${shape}§r brush has been made${blockData !== "defaultState" ?  " with block state [" + blockData + "]" : ""}`);

        this._firstTime = false;
    }

    setBlock(block, blockData) {
        this.block = block;
        this.blockData = BlockStateStringToObject(blockData) || {};

        this.setBlockPermutation(BlockPermutation.resolve(this.block, this.blockData), blockData == "defaultState");
    }

    setBlockPermutation(blockPerm, _hideState=false) {
        this.blockPerm = blockPerm;

        if (!this._firstTime) {
            const states = blockPerm.getAllStates();
            let hideState = _hideState || Object.keys(states).length == 0;
            notify(`Brush block changed to §9${blockPerm.type.id}§r ${!hideState ? "with block state [" + JSON.stringify(states) + "]" : ""}`);
        }
    }

    setMask(mask) {
        this.mask = mask;
        if (mask != null) {
            this.mask = mask.map(m => {
                return {
                    block: m.block,
                    blockData: BlockStateStringToObject(m.blockData)
                };
            });
        }
        
        if (!this._firstTime) {
            if (mask == null)
                notify(`Brush mask §9removed§r`);
            else {
                notify(`Brush mask changed to:\n§9${formatMaskArray(mask)}§r`);
            }
        }
    }


    setBlockPermToMask(blockPerm) {
        this.mask = [{
            block: blockPerm.type.id,
            blockData: blockPerm.getAllStates()
        }];

        if (!this._firstTime) {
            const blockData = this.mask[0].blockData;
            const formattedBlockData = Object.keys(blockData).map(key => `${key}=${blockData[key]}`);
            const tempMask = [{block: this.mask[0].block, blockData: formattedBlockData}]
            notify(`Brush mask changed to:\n§9${formatMaskArray(tempMask)}§r`);
        }

    }

    setSize(size) {
        this.size = parseInt(size);
        if (this.size > 20) {
            notify(`Brush size [${this.size}] is too big. Brush size set to 20`, "warn");
            this.size = 20;
        }

        if (!this._firstTime)
            notify(`Brush size changed to §2${this.size}§r`);
    }

    setShape(shape) {
        if (shape == "cyl") {
            this.shape = "cylinder";
            return;
        }
        else if (shape == "pat") {
            this.shape = "pattern"
            return;
        }
            
        this.shape = shape;
        
        if (!this._firstTime)
            notify(`Brush shape changed to §2${this.shape}§r`); 
    }
    

    build(pos, yAdd = 0) {
        if (this.shape == "cube") {
            this.cube(pos);
            return;
        }
        else if (this.shape == "pattern") {
            this.pattern(pos);
            return;
        }

        const startPos = pos;
        const endPos = Vector.add(pos, new Vector(0, yAdd - 0, 0));

        let rad = this.size;

        const regionsToFill = bakedShapes[this.shape][rad];
        for (const region of regionsToFill) {
            const fillStart = Vector.add(region.start, startPos);
            const fillEnd = Vector.add(region.end, endPos);

            if (this.mask == null) 
                fillBlocks(this.owner.dimension, fillStart, fillEnd, this.blockPerm);
            else
                fillBlocksMasked(this.owner.dimension, Vector.add(region.start, startPos), Vector.add(region.end, endPos), this.blockPerm, this.mask);
        }
    }

    cube(pos) {
        if (this.mask == null) {
            fillBlocks(this.owner.dimension,
                       Vector.subtract(pos, new Vector(this.size, this.size, this.size)),
                       Vector.add(pos, new Vector(this.size, this.size, this.size)),
                       this.blockPerm);
        }
        else {
            fillBlocksMasked(this.owner.dimension,
                             Vector.subtract(pos, new Vector(this.size, this.size, this.size)),
                             Vector.add(pos, new Vector(this.size, this.size, this.size)),
                             this.blockPerm,
                             this.mask);
        }
    }

    pattern(pos) {
        for (let j = 0; j < this.size * 7; j++) {
            let offset = new Vector(Math.floor((Math.random() - 0.5) * 2 * this.size),
                                    Math.floor((Math.random() - 0.5) * 2 * this.size),
                                    Math.floor((Math.random() - 0.5) * 2 * this.size));
  
            if (this.mask == null) {
                fillBlocks(this.owner.dimension,
                           Vector.add(pos, offset),
                           Vector.add(pos, offset),
                           this.blockPerm);
            }
            else {
                fillBlocksMasked(this.owner.dimension,
                                 Vector.add(pos, offset),
                                 Vector.add(pos, offset),
                                 this.blockPerm,
                                 this.mask);
            }
        }
    }

    attemptBuild(maxRange = 448) {
        const hit = this.owner.getBlockFromViewDirection({ maxDistance: maxRange });
        if (hit != undefined) {
            if (Vector.distance(this.owner.location, hit.block.location) < this.size + 1
            && this.mask == null
            && !world.getDynamicProperty("_allowBuildClipping")) {
                notify("You are too close to target, you will be buried. Move back a bit.", "warn");
                return;
            }

            this.build(Vector.add(hit.block.location, new Vector(0, 1, 0)));
        }
        else 
            notify("Target out of reach", "warn");

    }
}